import mpi4jax.cython  # noqa: F401

__all__ = ["Allreduce", "Send", "Recv", "Sendrecv"]

from .collective_ops.allreduce import Allreduce
from .collective_ops.send import Send
from .collective_ops.recv import Recv
from .collective_ops.sendrecv import Sendrecv

from .warn import disable_omnistaging_warning



from . import mpi_xla_bridge
from jax.lib import xla_client

import os

mpi_xla_bridge.set_logging(
    os.environ.get("MPI4JAX_DEBUG", "").lower() in ("true", "1", "on")
)

for name, fn in mpi_xla_bridge.cpu_custom_call_targets.items():
    xla_client.register_cpu_custom_call_target(name, fn)

del os, xla_client



import numpy as _np
import ctypes

from mpi4py import MPI as _MPI

from jax.lib import xla_client

_ops = xla_client.ops


def to_mpi_ptr(mpi_obj):
    """
    to_mpi_ptr(mpi_obj)

    Returns the ptr to the underlying C mpi object
    """
    return _np.uint64(_MPI._handleof(mpi_obj))


def MPIComm_from_ptr(ptr):
    """
    MPIComm_from_ptr(ptr)

    Constructs a MPI Comm object from a pointer
    """
    comm = _MPI.Comm()
    comm_ptr = ctypes.c_void_p.from_address(_MPI._addressof(comm))
    comm_ptr.value = int(ptr)
    return comm


def MPIOp_from_ptr(ptr):
    """
    MPIOp_from_ptr(ptr)

    Constructs a MPI Op object from a pointer
    """
    op = _MPI.Op()
    op_ptr = ctypes.c_void_p.from_address(_MPI._addressof(op))
    op_ptr.value = int(ptr)
    return op


def dtype_ptr(dtype):
    """
    dtype_ptr(dtype)

    Returns the pointer to the MPI dtype of the input numpy dtype
    """
    if dtype == _np.float32:
        _dtype = to_mpi_ptr(_MPI.FLOAT)
    elif dtype == _np.float64:
        _dtype = to_mpi_ptr(_MPI.DOUBLE)
    elif dtype == _np.complex64:
        _dtype = to_mpi_ptr(_MPI.COMPLEX)
    elif dtype == _np.complex128:
        _dtype = to_mpi_ptr(_MPI.DOUBLE_COMPLEX)
    elif dtype == _np.int32:
        _dtype = to_mpi_ptr(_MPI.INT32_T)
    elif dtype == _np.int64:
        _dtype = to_mpi_ptr(_MPI.INT64_T)

    return _dtype


# Helper functions
def _constant_s32_scalar(c, x):
    return _ops.Constant(c, _np.int32(x))


def _constant_s64_scalar(c, x):
    return _ops.Constant(c, _np.int64(x))


def _constant_u32_scalar(c, x):
    return _ops.Constant(c, _np.uint32(x))


def _constant_u64_scalar(c, x):
    return _ops.Constant(c, _np.uint64(x))


def _unpack_builder(c):
    # If `c` is a ComputationBuilder object, extracts the underlying XlaBuilder.
    return getattr(c, "_builder", c)



from setuptools import setup
from setuptools.extension import Extension
from Cython.Build import cythonize
import os


def mpi_info(cmd):
    import mpi4py

    config = mpi4py.get_config()
    cmd_compile = " ".join([config["mpicc"], "-show"])
    out_stream = os.popen(cmd_compile)
    flags = out_stream.read().strip()
    flags = flags.replace(",", " ").split()

    if cmd == "compile":
        startwith = "-I"
    elif cmd == "libdirs":
        startwith = "-L"
    elif cmd == "libs":
        startwith = "-l"

    out = []
    for flag in flags:
        if flag.startswith(startwith):
            out.append(flag[2:])
    return out


setup(
    name="mpi4jax",
    version="0.2.2",
    author="Filippo Vicentini",
    author_email="filippovicentini@gmail.com",
    long_description="""Jax-mpi provides integration among jax and MPI, so that
    code containing MPI calls can be correctly jit-compiled through jax.""",
    url="https://github.com/PhilipVinc/mpi4jax",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=["mpi4jax", "mpi4jax.collective_ops", "mpi4jax.cython"],
    ext_modules=cythonize(
        [
            Extension(
                name="mpi4jax.cython.mpi_xla_bridge",
                sources=["mpi4jax/cython/mpi_xla_bridge.pyx"],
                include_dirs=mpi_info("compile"),
                library_dirs=mpi_info("libdirs"),
                libraries=mpi_info("libs"),
            ),
        ],
    ),
    setup_requires=["setuptools>=18.0", "cython>=0.21", "mpi4py>=3.0.1"],
    python_requires=">=3.6",
    install_requires=["jax", "jaxlib>=0.1.55", "mpi4py>=3.0.1", "numpy"],
    extras_require={"dev": ["pytest"]},
)



import numpy as _np

from mpi4py import MPI as _MPI

from jax import abstract_arrays, device_put
from jax import numpy as jnp
from jax.lax import create_token
from jax.core import Primitive
from jax.lib import xla_client
from jax.interpreters import xla

from ..utils import (
    to_mpi_ptr,
    _unpack_builder,
    _ops,
    _constant_s32_scalar,
    _constant_u64_scalar,
    dtype_ptr,
)

from ..warn import warn_missing_omnistaging

# The Jax primitive
mpi_allreduce_p = Primitive("allreduce_mpi")  # Create the primitive


# This function applies the primitive to an AST
def Allreduce(x, op, comm=_MPI.COMM_WORLD, token=None):
    """
    Allreduce(x, op, comm=_MPI.COMM_WORLD, token=None)

    Performs the Allreduce operation `op` on the input `x` using the
    communicator `comm` which defaults to the world comunicator.
    An optional token can be passed, which is used to force jax to execute
    MPI operations in the correct order.

    Argumemnts:
        x: Array or scalar input.
        op: The reduction operation `MPI.Op` (e.g: MPI.SUM)
        comm: The communicator (defaults to MPI.COMM_WORLD)
        token: token to force a sequential order in the operations (default=None)

    Returns:
        res: result of the allreduce operation
        new_token: a new, modified token, that depends on this operation.
            This result can be ignored if result forces a data dependency.
    """
    if token is None:
        token = create_token(x)

    return mpi_allreduce_p.bind(x, token, op=op, comm=comm)


#  this function executes the primitive, when not under any transformation
def mpi_allreduce_impl(x, token, op, comm):
    # TODO: make this support gpus (use cupy?)
    inpt = _np.asarray(x)
    out = _np.zeros_like(x)

    comm.Allreduce(inpt, out, op=op)

    res = jnp.array(out, dtype=inpt.dtype)

    # if it's a jax array and not a standard python array
    if hasattr(x, "device_buffer"):
        # put the result on the correct device if needed
        if not (res.device_buffer.device() == x.device_buffer.device()):
            res = device_put(res, device=x.device_buffer.device())

    return res, token


#  This function compiles the operation
def mpi_allreduce_xla_encode(c, x, token, op, comm):
    warn_missing_omnistaging()

    c = _unpack_builder(c)
    x_shape = c.GetShape(x)
    dtype = x_shape.element_type()
    dims = x_shape.dimensions()

    # compute total number of elements in array
    _nitems = _constant_s32_scalar(c, _np.prod(dims, dtype=int))

    _dtype_ptr = dtype_ptr(dtype)

    sh = xla_client.Shape.tuple_shape(
        [xla_client.Shape.array_shape(dtype, dims), xla_client.Shape.token_shape()]
    )

    return _ops.CustomCall(
        c,
        b"mpi_allreduce",
        operands=(
            _nitems,
            x,
            _constant_u64_scalar(c, to_mpi_ptr(op)),
            _constant_u64_scalar(c, to_mpi_ptr(comm)),
            _constant_u64_scalar(c, _dtype_ptr),
            token,
        ),
        shape=sh,
        has_side_effect=True,
    )


# This function evaluates only the shapes during AST construction
def mpi_allreduce_abstract_eval(xs, token, op, comm):
    return (
        abstract_arrays.ShapedArray(xs.shape, xs.dtype),
        abstract_arrays.abstract_token,
    )


mpi_allreduce_p.multiple_results = True
mpi_allreduce_p.def_impl(mpi_allreduce_impl)
mpi_allreduce_p.def_abstract_eval(mpi_allreduce_abstract_eval)

# assign to the primitive the correct encoder
xla.backend_specific_translations["cpu"][mpi_allreduce_p] = mpi_allreduce_xla_encode



import numpy as _np

from mpi4py import MPI as _MPI

import jax.numpy as jnp
from jax import abstract_arrays, device_put
from jax.lax import create_token
from jax.core import Primitive
from jax.lib import xla_client
from jax.interpreters import xla

from ..utils import (
    to_mpi_ptr,
    _unpack_builder,
    _ops,
    _constant_s32_scalar,
    _constant_u64_scalar,
    dtype_ptr,
)

from ..warn import warn_missing_omnistaging

# The Jax primitive
mpi_sendrecv_p = Primitive("sendrecv_mpi")  # Create the primitive


# This function applies the primitive to an AST
def Sendrecv(
    sendbuf,
    recvbuf,
    source,
    dest,
    sendtag=0,
    recvtag=_MPI.ANY_TAG,
    comm=_MPI.COMM_WORLD,
    status=None,
    token=None,
):
    if token is None:
        token = create_token(sendbuf)

    return mpi_sendrecv_p.bind(
        sendbuf,
        recvbuf,
        token,
        source=source,
        dest=dest,
        sendtag=sendtag,
        recvtag=recvtag,
        comm=comm,
        status=status,
    )


#  this function executes the primitive, when not under any transformation
def mpi_sendrecv_impl(
    sendbuf, recvbuf, token, source, dest, sendtag, recvtag, comm, status
):
    # TODO: make this support gpus (use cupy?)
    inarr = _np.asarray(sendbuf)
    outarr = _np.empty_like(recvbuf)

    comm.Sendrecv(
        sendbuf=inarr,
        dest=dest,
        sendtag=sendtag,
        recvbuf=outarr,
        source=source,
        recvtag=recvtag,
        status=status,
    )

    if hasattr(recvbuf, "dtype"):
        dt = recvbuf.dtype
    else:
        # probably a scalar
        dt = _np.dtype(type(recvbuf))

    res = jnp.array(outarr, dtype=dt)

    # if it's a jax array and not a standard python array
    if hasattr(recvbuf, "device_buffer"):
        # put the result on the correct device if needed
        if not (res.device_buffer.device() == recvbuf.device_buffer.device()):
            res = device_put(res, device=recvbuf.device_buffer.device())

    return res, token


#  This function compiles the operation
def mpi_sendrecv_xla_encode(
    c, sendbuf, recvbuf, token, source, dest, sendtag, recvtag, comm, status
):
    from ..cython.mpi_xla_bridge import MPI_STATUS_IGNORE_ADDR

    warn_missing_omnistaging()

    c = _unpack_builder(c)

    recv_shape = c.GetShape(recvbuf)
    recv_dtype = recv_shape.element_type()
    recv_dims = recv_shape.dimensions()

    # compute total number of elements in array
    _recv_nitems = _constant_s32_scalar(c, _np.prod(recv_dims, dtype=int))
    _recv_dtype_ptr = dtype_ptr(recv_dtype)

    send_shape = c.GetShape(sendbuf)
    send_dtype = send_shape.element_type()
    send_dims = send_shape.dimensions()

    # compute total number of elements in array
    _send_nitems = _constant_s32_scalar(c, _np.prod(send_dims, dtype=int))
    _send_dtype_ptr = dtype_ptr(send_dtype)

    sh = xla_client.Shape.tuple_shape(
        [
            xla_client.Shape.array_shape(recv_dtype, recv_dims),
            xla_client.Shape.token_shape(),
        ]
    )

    if status is None:
        _status = MPI_STATUS_IGNORE_ADDR
    else:
        _status = _MPI._addressof(status)

    operands = (
        _send_nitems,
        sendbuf,
        _constant_s32_scalar(c, dest),
        _constant_s32_scalar(c, sendtag),
        _constant_u64_scalar(c, _send_dtype_ptr),
        _recv_nitems,
        _constant_s32_scalar(c, source),
        _constant_s32_scalar(c, recvtag),
        _constant_u64_scalar(c, _recv_dtype_ptr),
        _constant_u64_scalar(c, to_mpi_ptr(comm)),
        _constant_u64_scalar(c, _status),
        token,
    )

    return _ops.CustomCall(
        c,
        b"mpi_sendrecv",
        operands=operands,
        shape=sh,
        has_side_effect=True,
    )


# This function evaluates only the shapes during AST construction
def mpi_sendrecv_abstract_eval(
    sendbuf, recvbuf, token, source, dest, sendtag, recvtag, comm, status
):
    return (
        abstract_arrays.ShapedArray(recvbuf.shape, recvbuf.dtype),
        abstract_arrays.abstract_token,
    )


mpi_sendrecv_p.multiple_results = True
mpi_sendrecv_p.def_impl(mpi_sendrecv_impl)
mpi_sendrecv_p.def_abstract_eval(mpi_sendrecv_abstract_eval)

# assign to the primitive the correct encoder
xla.backend_specific_translations["cpu"][mpi_sendrecv_p] = mpi_sendrecv_xla_encode



import numpy as _np

from mpi4py import MPI as _MPI

from jax import abstract_arrays
from jax.lax import create_token
from jax.core import Primitive
from jax.lib import xla_client
from jax.interpreters import xla

from ..utils import (
    to_mpi_ptr,
    _unpack_builder,
    _ops,
    _constant_s32_scalar,
    _constant_u64_scalar,
    dtype_ptr,
)

from ..warn import warn_missing_omnistaging


# The Jax primitive
mpi_send_p = Primitive("send_mpi")  # Create the primitive


# This function applies the primitive to an AST
def Send(x, dest, tag=0, comm=_MPI.COMM_WORLD, token=None):
    """
    Send(x, dest, tag=0, comm=_MPI.COMM_WORLD, token=None)

    Sends the input`x` to the target rank `dest` using the communicator `comm`
    which defaults to the  world comunicator, with the `tag`.
    An optional token can be passed, which is used to force jax to execute
    MPI operations in the correct order.
    This is particularly important if you are performing different Send/Recv
    operations, which might otherwise deadlock.

    Argumemnts:
        x: Array or scalar input.
        dest: rank of the target MPI process.
        tag: Tag of this message.
        comm: The communicator (defaults to MPI.COMM_WORLD)
        token: token to force a sequential order in the operations (default=None)

    Returns:
        new_token: a new, modified token, that depends on this operation.
    """
    if token is None:
        token = create_token(x)

    out = mpi_send_p.bind(x, token, dest=dest, tag=tag, comm=comm)
    return out


#  this function executes the primitive, when not under any transformation
def mpi_send_impl(x, token, dest, tag, comm):
    # TODO: make this support gpus (use cupy?)
    inpt = _np.asarray(x)
    comm.Send(inpt, dest=dest, tag=tag)
    return token


#  This function compiles the operation
def mpi_send_xla_encode(c, x, token, dest, tag, comm):
    warn_missing_omnistaging()

    c = _unpack_builder(c)
    x_shape = c.GetShape(x)
    dtype = x_shape.element_type()
    dims = x_shape.dimensions()

    # compute total number of elements in array
    _nitems = _constant_s32_scalar(c, _np.prod(dims, dtype=int))
    _dtype_ptr = dtype_ptr(dtype)

    # ensure void** out type
    sh = xla_client.Shape.tuple_shape([xla_client.Shape.token_shape()])

    out = _ops.CustomCall(
        c,
        b"mpi_send",
        operands=(
            _nitems,
            x,
            _constant_s32_scalar(c, dest),
            _constant_s32_scalar(c, tag),
            _constant_u64_scalar(c, to_mpi_ptr(comm)),
            _constant_u64_scalar(c, _dtype_ptr),
            token,
        ),
        shape=sh,
        has_side_effect=True,
    )

    return xla_client.ops.GetTupleElement(out, 0)


# This function evaluates only the shapes during AST construction
def mpi_send_abstract_eval(xs, token, dest, tag, comm):
    return abstract_arrays.abstract_token


# mpi_send_p.multiple_results = True
mpi_send_p.def_impl(mpi_send_impl)
mpi_send_p.def_abstract_eval(mpi_send_abstract_eval)

# assign to the primitive the correct encoder
xla.backend_specific_translations["cpu"][mpi_send_p] = mpi_send_xla_encode
